import importlib
import importlib.util
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    ForwardRef,
    Generic,
    TypeVar,
    Union,
    cast,
)

TypeName = TypeVar("TypeName")
Module = TypeVar("Module")
Other = TypeVar("Other")


@dataclass(frozen=True)
class LazyType(Generic[TypeName, Module]):
    """A class that represents a type that will be resolved at runtime.

    This is useful when you have circular dependencies between types.

    This class is not meant to be used directly, instead use the `strawberry.lazy`
    function.
    """

    type_name: str
    module: str
    package: str | None = None

    def __or__(self, other: Other) -> object:
        return Union[self, other]  # noqa: UP007

    def resolve_type(self) -> type[Any]:
        module = importlib.import_module(self.module, self.package)

        # Resolve full module name for __main__ comparison
        if self.package:
            full_module_name = importlib.util.resolve_name(self.module, self.package)
        else:
            full_module_name = self.module

        main_module = sys.modules.get("__main__", None)
        if main_module:
            # If lazy type points to the main module, use it instead of the imported
            # module. Otherwise duplication checks during schema-conversion might fail.
            # Refer to: https://github.com/strawberry-graphql/strawberry/issues/2397
            if main_module.__spec__ and main_module.__spec__.name == full_module_name:
                module = main_module
            elif hasattr(main_module, "__file__") and hasattr(module, "__file__"):
                main_file = main_module.__file__
                module_file = module.__file__
                if main_file and module_file:
                    try:
                        is_samefile = Path(main_file).samefile(module_file)
                    except FileNotFoundError:
                        # Can be raised when run through the CLI as the __main__ file
                        # path contains `strawberry.exe`
                        is_samefile = False
                    module = main_module if is_samefile else module

        return module.__dict__[self.type_name]

    def __call__(self) -> None:  # pragma: no cover
        # this empty call method allows LazyTypes to be used in generic types
        return None


class StrawberryLazyReference:
    """A class that represents a lazy reference to a type in another module.

    This is useful when you have circular dependencies between types.

    This class is not meant to be used directly, instead use the `strawberry.lazy`
    function.
    """

    def __init__(self, module: str) -> None:
        self.module = module
        self.package = None

        if module.startswith("."):
            frame = sys._getframe(2)
            # TODO: raise a nice error if frame is None
            assert frame is not None
            self.package = cast("str", frame.f_globals["__package__"])

    def resolve_forward_ref(self, forward_ref: ForwardRef) -> LazyType:
        return LazyType(forward_ref.__forward_arg__, self.module, self.package)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, StrawberryLazyReference):
            return NotImplemented

        return self.module == other.module and self.package == other.package

    def __hash__(self) -> int:
        return hash((self.__class__, self.module, self.package))


def lazy(module_path: str) -> StrawberryLazyReference:
    """Creates a lazy reference to a type in another module.

    Args:
        module_path: The path to the module containing the type, supports relative paths
            starting with `.`

    Returns:
        A `StrawberryLazyReference` object that can be used to reference a type in another
        module.

    This is useful when you have circular dependencies between types.

    For example, assuming you have a `Post` type that has a field `author` that
    references a `User` type (which also has a field `posts` that references a list of
    `Post`), you can use `strawberry.lazy` to avoid the circular dependency:

    ```python
    from typing import TYPE_CHECKING, Annotated

    import strawberry

    if TYPE_CHECKING:
        from .users import User


    @strawberry.type
    class Post:
        title: str
        author: Annotated["User", strawberry.lazy(".users")]
    ```
    """
    return StrawberryLazyReference(module_path)


__all__ = ["LazyType", "StrawberryLazyReference", "lazy"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Literal
from typing_extensions import TypedDict

from strawberry.schema._graphql_core import (
    GraphQLIncrementalExecutionResults,
    ResultType,
)


class GraphQLHTTPResponse(TypedDict, total=False):
    data: dict[str, object] | None
    errors: list[object] | None
    extensions: dict[str, object] | None
    hasNext: bool | None
    completed: list[Any] | None
    pending: list[Any] | None
    initial: list[Any] | None
    incremental: list[Any] | None


def process_result(result: ResultType) -> GraphQLHTTPResponse:
    if isinstance(result, GraphQLIncrementalExecutionResults):
        return result

    errors, extensions = result.errors, result.extensions
    data: GraphQLHTTPResponse = {
        "data": result.data,
        **({"errors": [err.formatted for err in errors]} if errors else {}),
        **({"extensions": extensions} if extensions else {}),
    }

    return data


GraphQLRequestProtocol = Literal["http", "multipart-subscription", "sse"]


@dataclass
class GraphQLRequestData:
    # query is optional here as it can be added by an extensions
    # (for example an extension for persisted queries)
    query: str | None
    variables: dict[str, Any] | None
    operation_name: str | None
    extensions: dict[str, Any] | None
    protocol: GraphQLRequestProtocol = "http"


__all__ = [
    "GraphQLHTTPResponse",
    "GraphQLRequestData",
    "GraphQLRequestProtocol",
    "process_result",
]



from typing import Any

from strawberry.test import BaseGraphQLTestClient


class GraphQLTestClient(BaseGraphQLTestClient):
    def request(
        self,
        body: dict[str, object],
        headers: dict[str, object] | None = None,
        files: dict[str, object] | None = None,
    ) -> Any:
        if files:
            return self._client.post(
                self.url, data=body, format="multipart", headers=headers
            )

        return self._client.post(
            self.url, data=body, content_type="application/json", headers=headers
        )


__all__ = ["GraphQLTestClient"]



from enum import Enum
from typing import NewType

from strawberry.types.enum import enum
from strawberry.types.scalar import scalar

FieldSet = NewType("FieldSet", str)
"""Represents a selection set for federation @requires, @provides, @key directives."""
FieldSet._scalar_definition = scalar(  # type: ignore[attr-defined]
    name="_FieldSet",
    serialize=lambda value: value,
    parse_value=str,
    print_definition=False,
)

LinkImport = NewType("LinkImport", object)
"""Represents an import for the @link directive."""
LinkImport._scalar_definition = scalar(  # type: ignore[attr-defined]
    name="link__Import",
    serialize=lambda value: value,
    parse_value=lambda value: value,
    print_definition=False,
)


@enum(name="link__Purpose", print_definition=False)
class LinkPurpose(Enum):
    SECURITY = "SECURITY"
    EXECUTION = "EXECUTION"


__all__ = ["FieldSet", "LinkImport", "LinkPurpose"]
